# -*- coding: utf-8 -*-
import os

from unipath import Path

from mootdx.financial import financial
from mootdx.logger import log
from mootdx.utils import TqdmUpTo


class Affair(object):
    @staticmethod
    def parse(downdir='.', filename=None):
        """
        按目录解析文件

        :param downdir:
        :param filename:
        :return:
        """

        if not filename:
            log.critical('文件名不能为空!')
            return None

        filepath = Path(downdir, filename)

        if Path(filepath).exists():
            return financial.FinancialReader().to_data(filepath)

        log.error('文件不存在：{}'.format(filename))

        return None

    @staticmethod
    def files():
        """
        财务文件列表

        :return:
        """

        history = financial.FinancialList()
        results = history.fetch_and_parse()

        return results

    @staticmethod
    def fetch(downdir='.', filename=None, *args, **kwargs):
        """
        财务数据下载

        :param downdir:
        :param filename:
        :param kwargs:
        :return:
        """

        history = financial.FinancialList()
        crawler = financial.Financial()

        if not os.path.isdir(downdir):
            log.warning('下载目录不存在, 进行创建.')
            os.makedirs(downdir)

        if filename:
            log.info('下载文件 {}.'.format(filename))
            downfile = os.path.join(downdir, filename)

            with TqdmUpTo(unit='B', unit_scale=True, miniters=1, ascii=True) as t:
                crawler.fetch_and_parse(report_hook=t.update_to, filename=filename, downdir=downfile)

            return True

        list_data = history.fetch_and_parse()

        for x in list_data:
            downfile = os.path.join(downdir, x['filename'])

            # 判断文件存在并且长度一样，则忽略
            if os.path.exists(downfile):
                if int(x.get('filesize')) == int(os.path.getsize(downfile)):
                    log.warning('[!] 文件已经存在: {} 跳过.'.format(x['filename']))
                    continue

            with TqdmUpTo(unit='b', unit_scale=True, miniters=1, ascii=True) as t:
                print('\r[+] 准备下载文件 {}.'.format(x['filename']))
                crawler.fetch_and_parse(report_hook=t.update_to, filename=x['filename'], downdir=downfile)



# -*- coding: utf-8 -*-
"""
用于管理缓存的配置数据
使用前必须先调用 init() 。
"""
import copy
import json

from unipath import Path

from mootdx.consts import EX_HOSTS, GP_HOSTS, HQ_HOSTS
from mootdx.logger import log
from mootdx.server import bestip
from mootdx.utils import get_config_path

__all__ = ['set', 'get', 'copy', 'update', 'settings']

settings = {
    'SERVER': {
        'HQ': HQ_HOSTS,
        'EX': EX_HOSTS,
        'GP': GP_HOSTS
    },
    'BESTIP': {
        'HQ': '',
        'EX': '',
        'GP': ''
    },
    'TDXDIR': 'C:/new_tdx',
}

BASE = Path(__file__).parent.parent
CONF = get_config_path('config.json')


def setup():
    """
    将 yaml 里的配置文件导入到 config.py 中
    :return: bool，true 表示数据导入成功。
    """
    global settings

    try:
        options = json.load(open(CONF))
        settings.update(options)
    except Exception as e:
        log.error(e)
        log.error(f'未找到配置文件 {CONF}, 正在生成配置文件.')
        bestip()

    return True if settings else False


def has(key, value):
    """
    通过 key 设置某一项值

    :param key:
    :param value:
    :return:
    """

    return value in settings[key]


def set(key, value):
    """
    通过 key 设置某一项值

    :param key:
    :param value:
    :return:
    """

    settings[key] = value


def get(key, default=None):
    """
    通过 key 获取值

    :param key:
    :param default:
    :return:
    """

    key = key.split('.')
    cfg = settings.get(key[0])

    if len(key) > 1:
        for x in key[1:]:
            if cfg.get(x):
                cfg = cfg.get(x)
            else:
                cfg = cfg.get(x, default)
                break

    return cfg


def path(key, value=None):
    """
    通过 key 构建路径

    :param key:
    :param value:
    :return:
    """

    return Path(BASE, settings.get(key), value)


def clone():
    """
    复制配置

    :return:
    """

    return copy.deepcopy(settings)


def update(options):
    """
    全部替换配置

    :param options:
    :return:
    """

    settings.update(options)




import json
import re

import pandas as pd
import requests
import requests.exceptions
from numpy import asarray

from mootdx.logger import log


def get_adjust_year(symbol=None, year='2021', factor='00'):
    """
    采集同花顺复权数据

    :param symbol: 股票代码
    :param factor: 前后复权 before 或 01 为前复权， after 或 02为后复权
    :param year: 年份
    :return: DataFrame
    """

    # http://d.10jqka.com.cn/v2/line/hs_600036/01/2018.js
    # http://d.10jqka.com.cn/v6/line/hs_600000/00/all.js

    headers = {
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.212 Safari/537.36',
        'Referer': 'http://stockpage.10jqka.com.cn/',
        'DNT': '1',
    }

    if factor == 'before':
        factor = '01'
    elif factor == 'after':
        factor = '02'

    if factor not in ['01', '02']:
        log.warning('复权参数错误，factor 的值必须是: before, after, 01, 02')
        return pd.DataFrame(data=None)

    try:
        url = f'http://d.10jqka.com.cn/v2/line/hs_{symbol}/{factor}/{year}.js'
        res = requests.get(url, headers=headers)
        res.raise_for_status()

        # 出现 window.location.href 则请求太频繁，需要稍等再采集
        text = re.findall(r'\((.*)\)', res.text)[0]
        text = json.loads(text)

        data = text['data'].split(';')
        data = [item.split(',')[:8] for item in data]

        columns = ['date', 'open', 'high', 'low', 'close', 'volume', 'amount', 'adjust']
        return pd.DataFrame(data, index=list(asarray(data).T[0]), columns=columns)
    except requests.exceptions.RequestException:
        log.warning('网络连接失败，请重试...')
    except IndexError as e:
        log.warning('数据解析错误，请求太频繁，请稍后重试...')
        log.debug(e)

    return pd.DataFrame(data=None)



# -*- coding: utf-8 -*-
import json
import logging
import os

import click
from prettytable import PrettyTable

from mootdx import CONFIG, __version__
from mootdx.affair import Affair
from mootdx.logger import log
from mootdx.quotes import Quotes
from mootdx.reader import Reader
from mootdx.server import Server
from mootdx.utils import get_config_path, to_file


@click.group()
@click.option('--debug/--no-debug', default=False)
@click.pass_context
def cli(ctx, debug):
    ctx.obj["DEBUG"] = debug
    click.echo(f"Debug mode is {'on' if debug else 'off'}")

    if debug:
        logging.basicConfig(level=logging.DEBUG)


@cli.command(help='读取股票在线行情数据.')
@click.option('-o', '--output', default=None, help='输出文件, 支持CSV, HDF5, Excel等格式.')
@click.option('-s', '--symbol', default='600000', help='股票代码.')
@click.option('-a', '--action', default='bars', help='操作类型 (daily: 日线, minute: 一分钟线, fzline: 五分钟线).')
@click.option('-m', '--market', default='std', help='证券市场, 默认 std (std: 标准股票市场, ext: 扩展市场).')
def quotes(symbol, action, market, output):
    client = Quotes.factory(market=market, multithread=True, heartbeat=True)

    try:
        action = 'bars' if 'daily' else action
        if action == 'daily':
            frequency = 9
        elif action == 'minute':
            frequency = 8
        elif action == 'fzline':
            frequency = 0
        else:
            frequency = 9

        feed = getattr(client, 'bars')(symbol=symbol, frequency=frequency)
        to_file(feed, output) if output else None
        print(feed)
    except Exception as e:
        raise e


@cli.command(help='读取股票本地行情数据.')
@click.option('-d', '--tdxdir', default='C:/new_tdx', help='通达信数据目录.')
@click.option('-s', '--symbol', default='600000', help='股票代码.')
@click.option('-a', '--action', default='daily', help='操作类型 (daily: 日线, minute: 一分钟线, fzline: 五分钟线).')
@click.option('-m', '--market', default='std', help='证券市场, 默认 std (std: 标准股票市场, ext: 扩展市场).')
@click.option('-o', '--output', default=None, help='输出文件, 支持 CSV, HDF5, Excel 等格式.')
def reader(symbol, action, market, tdxdir, output):
    client = Reader.factory(market=market, tdxdir=tdxdir)

    try:
        feed = getattr(client, action)(symbol=symbol)
        to_file(feed, output) if output else None
        print(feed)
    except Exception as e:
        raise e


@cli.command(help='测试行情服务器.')
@click.option('-l', '--limit', default=5, help='显示最快前几个，默认 5.')
@click.option('-w', '--write', count=True, help='将最优服务器IP写入配置文件 ~/.mootdx/config.json.')
@click.option('-v', '--verbose', count=True)
def bestip(limit, write, verbose):
    """
    @todo 命令行最优线路配置功能调整
    :param limit:
    :param write:
    :param verbose:
    :return:
    """
    config = get_config_path('config.json')
    default = CONFIG

    for index in ['HQ', 'EX', 'GP']:
        result = Server(index=index, limit=limit, console=True, verbose=verbose)

        if result:
            default['BESTIP'][index] = result[0]
        else:
            print(result)

    if write:
        json.dump(CONFIG, open(config, 'w'), indent=2)
        print('[√] 已经将最优服务器IP写入配置文件 {}'.format(config))


@cli.command(help='财务文件下载&解析.')
@click.option('-p', '--parse', default=None, help='要解析文件名')
@click.option('-l', '--files', count=True, default=True, help='列出文件列表')
@click.option('-f', '--fetch', default=None, help='下载财务文件的文件名')
@click.option('-a', '--downall', count=True, help='下载全部文件')
@click.option('-o', '--output', default=None, help='输出文件, 支持 CSV, HDF5, Excel, JSON 等格式.')
@click.option('-d', '--downdir', default='output', help='下载文件目录')
@click.option('-v', '--verbose', count=True)
def affair(parse, files, fetch, downdir, output, downall, verbose):
    if verbose:
        logging.basicConfig(level=logging.DEBUG)

    result = Affair.files()

    if not fetch and not parse:
        t = PrettyTable(["filename", "filesize", "hash"])
        t.align["filename"] = "l"
        t.align["filesize"] = "l"
        t.align["hash"] = "l"
        t.padding_width = 0

        for x in result:
            t.add_row([x['filename'], x['filesize'], x['hash']])

        print(t)

    if downall or fetch == 'all':
        feed = Affair.fetch(downdir=downdir)
        to_file(feed, output) if output else None
    elif fetch:
        Affair.fetch(downdir=downdir, filename=fetch.strip('.zip') + '.zip')

    if parse:
        files = [x['filename'] for x in result]

        if parse in files:
            if not os.path.exists(os.path.join(downdir, parse)):
                Affair.fetch(downdir=downdir, filename=parse.strip('.zip') + '.zip')

            feed = Affair.parse(downdir=downdir, filename=parse.strip('.zip') + '.zip')

            if output:
                to_file(feed, output)
            else:
                print(feed)
        else:
            log.error('cannot find file.')


@cli.command(help='显示当前软件版本.')
def version():
    print('mootdx v{}'.format(__version__))


@cli.command(help='批量下载行情数据.')
@click.option('-o', '--output', default='bundle', help='转存文件目录.')
@click.option('-s', '--symbol', default='600000', help='股票代码. 多个用,隔开')
@click.option('-a', '--action', default='bars', help='操作类型 (daily: 日线, minute: 一分钟线, fzline: 五分钟线).')
@click.option('-m', '--market', default='std', help='证券市场, 默认 std (std: 标准股票市场, ext: 扩展市场).')
@click.option('-e', '--extension', default='csv', help='转存文件的格式, 支持 CSV, HDF5, Excel, JSON 等格式.')
def bundle(symbol, action, market, output, extension):
    """
    批量下载行情数据
    :return:
    """
    client = Quotes.factory(market=market)
    symbol = symbol.replace('，', ',').strip(',').split(',')

    for code in symbol:
        try:
            if action == 'daily':
                frequency = 9
            elif action == 'minute':
                frequency = 8
            elif action == 'fzline':
                frequency = 0
            else:
                frequency = 9

            feed = getattr(client, 'bars')(symbol=code, frequency=frequency)
            to_file(feed, os.path.join(output, f'{code}.{extension}')) if output else None
            print('下载完成 {}'.format(code))
        except Exception as e:
            raise e

    print('[√] 下载文件到 "{}"'.format(os.path.realpath(output)))


def entry():
    cli(obj={})


if __name__ == "__main__":
    entry()



# -*- coding: utf-8 -*-

from mootdx import config
from mootdx.consts import EX_HOSTS, GP_HOSTS, HQ_HOSTS
from mootdx.server import Server
from mootdx.utils import get_config_path

__version__ = '0.7.5'
__author__ = "bopo.wang<ibopo@126.com>"

config.setup()



# -*- coding: utf-8 -*-
import hashlib
import os
import platform
from datetime import datetime
from struct import calcsize, unpack

import pandas as pd
from pandas import DataFrame
from tqdm import tqdm
from unipath import Path

from mootdx.consts import MARKET_SH, MARKET_SZ
from mootdx.logger import log


def get_stock_markets(symbols=None):
    results = []

    assert isinstance(symbols, list), 'stock code need list type'

    if isinstance(symbols, list):
        for symbol in symbols:
            results.append([get_stock_market(symbol, string=False), symbol.strip('sh').strip('sz')])

    return results


def get_stock_market(symbol='', string=False):
    """
    判断股票ID对应的证券市场匹配规则

    ['50', '51', '60', '90', '110'] 为 sh
    ['00', '12'，'13', '18', '15', '16', '18', '20', '30', '39', '115'] 为 sz
    ['5', '6', '9'] 开头的为 sh， 其余为 sz

    :param string: False 返回市场ID，否则市场缩写名称
    :param symbol: 股票ID, 若以 'sz', 'sh' 开头直接返回对应类型，否则使用内置规则判断
    :return 'sh' or 'sz'
    """

    assert isinstance(symbol, str), 'stock code need str type'

    market = None

    if symbol.startswith(('sh', 'sz')):
        market = symbol[:2]

    elif symbol.startswith(('50', '51', '60', '68', '90', '110', '113', '132', '204')):
        market = 'sh'

    elif symbol.startswith(('00', '12', '13', '18', '15', '16', '18', '20', '30', '39', '115', '1318')):
        market = 'sz'

    elif symbol.startswith(('5', '6', '9', '7')):
        market = 'sh'

    if string is False:
        market = MARKET_SZ if market == 'sz' else MARKET_SH

    return market


def gpcw(filepath):
    cw_file = open(filepath, 'rb')

    header_size = calcsize("<3h1H3L")
    stock_item_size = calcsize("<6s1c1L")

    data_header = cw_file.read(header_size)
    stock_header = unpack("<3h1H3L", data_header)

    max_count = stock_header[3]

    for idx in range(0, max_count):
        cw_file.seek(header_size + idx * calcsize("<6s1c1L"))
        si = cw_file.read(stock_item_size)
        stock_item = unpack("<6s1c1L", si)
        code = stock_item[0].decode()
        foa = stock_item[2]
        cw_file.seek(foa)
        info_data = cw_file.read(calcsize('<264f'))
        cw_info = unpack('<264f', info_data)

        log.debug(f"{code}, {cw_info}")
        return code, cw_info


def md5sum(downfile):
    """
    文件的 md5 哈希值

    :param downfile: 文件路径
    :return: mixed
    """

    try:
        md5_l = hashlib.md5()
        md5_l.update(Path(downfile).read_file('rb'))
        return md5_l.hexdigest()
    except (IOError, FileNotFoundError) as e:
        log.error(f'无法读取文件: {downfile}')
        log.debug(e)
        return None


def to_data(v):
    """
    数值转换为 pd.DataFrame

    :param v: mixed
    :return: pd.DataFrame
    """

    if not v:
        return pd.DataFrame(data=[])

    if isinstance(v, DataFrame):
        return v
    elif isinstance(v, list):
        return pd.DataFrame(data=v) if len(v) else None
    elif isinstance(v, dict):
        return pd.DataFrame(data=[v])

    return pd.DataFrame(data=[])


def to_file(df, filename=None):
    """
    根据扩展名输出文件

    :param df: pd.DataFrame
    :param filename: 要输出的文件，支持 csv, xlsx, xls, json, h5
    :return: bool
    """
    if filename is None or df is None:
        return None

    path_name = os.path.dirname(filename)

    if path_name and not os.path.isdir(path_name):
        os.makedirs(path_name)

    extension = os.path.splitext(filename)
    extension = extension[1] if len(extension) >= 2 else ''

    if extension == '.csv':
        return df.to_csv(filename, encoding='utf-8', index=False)
    elif extension == '.xlsx' or extension == '.xls':
        # openpyxl, xlwt
        return df.to_excel(filename, index=False)
    elif extension == '.h5':
        # tables
        return df.to_hdf(filename, 'df', index=False)
    elif extension == '.json':
        return df.to_json(filename, orient='records')

    return None


class TqdmUpTo(tqdm):
    """
    Provides `update_to(n)` which uses `tqdm.update(delta_n)`.

    """

    total: object = 0

    def update_to(self, downloaded=0, total_size=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if total_size is not None:
            self.total = total_size

        self.update(downloaded - self.n)  # will also set self.n = b * bsize


def get_config_path(config='config.json'):
    """
    获取配置文件路径

    :param config: 配置文件名称
    :return: filename
    """

    sub_path = 'mootdx' if platform.system() == 'Windows' else '.mootdx'
    filename = os.path.join(os.path.expanduser('~'), sub_path, config)
    pathname = os.path.dirname(filename)

    Path(pathname).mkdir(parents=True)

    return filename


def block_new(tdxdir=None, name: str = None, symbol: list = None):
    """
    自定义模块写入函数

    :param tdxdir: tdx 路径
    :param name: 自定义板块名称
    :param symbol: 自定义板块股票代码集合
    :return: bool
    """

    if not tdxdir:
        return False

    # 自定义板块名称未传入则自动按时间生成名称
    if not name:
        name = datetime.now().strftime('%Y%m%d%H%M%S')

    # 按时间生成 blk 文件名
    file = datetime.now().strftime('%Y%m%d%H%M%S')

    vipdoc = Path(tdxdir, 'T0002', 'blocknew')
    symbol = list(set(symbol))

    # 判断目录是否存在
    if not Path(vipdoc).isdir():
        log.error(f'自定义板块目录错误: {vipdoc}')
        return False

    block_file = Path(vipdoc, 'blocknew.cfg')

    # 文件不存在就创建
    if not block_file.exists():
        block_file.write_file('')

    # 判断名字是否重名
    with open(block_file, 'rb') as fp:
        names = fp.read().decode('gbk', 'ignore')
        names = names.split("\x00")
        names = [x for x in names if x != '']
        names = [v for i, v in enumerate(names) if i % 2 == 0]

        if name in names:
            log.error('自定义板块名称重复.')
            raise Exception('自定义板块名称重复.')

    # 写 blk 文件
    with open(f'{vipdoc}/{file}.blk', 'w') as fp:
        fp.write('\n'.join(symbol))

    # 写 blocknew.cfg 文件
    with open(block_file, 'ab') as fp:
        data = name + ((50 - len(name.encode('gbk', 'ignore'))) * "\x00")
        data += file + ((70 - len(file.encode('gbk', 'ignore'))) * "\x00")
        data = bytes(data.encode('gbk', 'ignore'))
        fp.write(data)

    return True



# -*- coding: utf-8 -*-
import struct
import tempfile
from urllib.request import Request, urlopen


def reporthook(downloaded, total_size):
    print("Downloaded {}, Total is {}".format(downloaded, total_size))


class BaseReader(object):
    @staticmethod
    def unpack(fmt, data):
        """
        解压文件

        :param fmt:
        :param data:
        :return:
        """

        record = struct.Struct(fmt)
        return (record.unpack_from(data, offset) for offset in range(0, len(data), record.size))

    def get_df(self, code_or_file, exchange=None):
        """
        转换格式为 pd.DateFrame

        :param code_or_file:
        :param exchange:
        :return:
        """

        raise NotImplementedError('not yet')


class BaseFinancial:
    def __init__(self, mode="content"):
        self.mode = mode

    def fetch_and_parse(self, report_hook=None, downdir=None, proxies=None, chunk_size=51200, *args, **kwargs):
        """
        function to get data , 参考 https://docs.python.org/3/library/urllib.request.html#module-urllib.request

        :param report_hook 使用urllib.request 的report_hook 来汇报下载进度
        :param downdir 数据文件下载的地址，如果没有提供，则下载到临时文件中，并在解析之后删除
        :param proxies urllib格式的代理服务器设置
        :param chunk_size chunk_size
        :return: 解析之后的数据结果
        """

        if self.mode == "http":
            file = self.fetch_via_http(report_hook=report_hook, downdir=downdir, proxies=proxies, chunk_size=chunk_size, *args, **kwargs)
        elif self.mode == "content":
            file = self.content(report_hook=report_hook, downdir=downdir, chunk_size=chunk_size, *args, **kwargs)
        else:
            return

        result = self.parse(file, *args, **kwargs)

        try:
            file.close()
        except Exception as e:
            raise e

        return result

    def fetch_via_http(self, report_hook=None, downdir=None, proxies=None, chunk_size=1024 * 50, *args, **kwargs):
        """

        :param report_hook:
        :param downdir:
        :param proxies:
        :param chunk_size:
        :param args:
        :param kwargs:
        :return:
        """

        download_file = open(downdir, 'wb') if downdir else tempfile.NamedTemporaryFile(delete=True)

        url = self.build_url(*args, **kwargs)

        request = Request(url)
        request.add_header('Referer', url)
        request.add_header(
            "User-Agent",
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_0) AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/76.0.3809.100 Safari/537.36 "
        )

        try:
            res = urlopen(request)
        except Exception as e:
            raise e

        if res.getheader('Content-Length') is not None:
            total_size = int(res.getheader('Content-Length').strip())
            downloaded = 0

            while True:
                chunk = res.read(chunk_size)
                downloaded += len(chunk)

                if report_hook is not None:
                    report_hook(downloaded, total_size)

                if not chunk:
                    break

                download_file.write(chunk)
        else:
            content = res.read()
            download_file.write(content)

        download_file.seek(0)
        return download_file

    def build_url(self, *args, **kwargs):
        raise NotImplementedError("will impl in subclass")

    def content(self, report_hook=None, downdir=None, proxies=None, chunk_size=1024 * 50, *args, **kwargs):
        raise NotImplementedError("will impl in subclass")

    def parse(self, download_file, *args, **kwargs):
        raise NotImplementedError("will impl in subclass")



# -*- coding: utf-8 -*-
import os
import random
import shutil
import tempfile
from abc import ABC
from struct import calcsize, unpack

import pandas as pd
from pytdx.hq import TdxHq_API
from unipath import Path

from mootdx import config
from mootdx.config import settings

from .base import BaseFinancial, BaseReader


class FinancialReader(BaseReader, ABC):
    @staticmethod
    def to_data(filename):
        """
        读取历史财务数据文件，并返回pandas结果 ， 类似 `gpcw20171231.zip` 格式，具体字段含义参考

        https://github.com/rainx/pytdx/issues/133

        :param filename: 数据文件地址， 数据文件类型可以为 .zip 文件，也可以为解压后的 .dat, 可以不写扩展名. 程序自动识别
        :return: pandas DataFrame 格式的历史财务数据
        """

        crawler = Financial()

        with open(filename, 'rb') as fp:
            data = crawler.parse(download_file=fp)

        return crawler.to_df(data)


class FinancialList(BaseFinancial):

    def build_url(self, *args, **kwargs):
        """
        获取采集数据的 URL

        :param args:
        :param kwargs:
        :return:
        """
        return "https://gitee.com/yutiansut/QADATA/raw/master/financial/content.txt"

    def content(self, report_hook=None, downdir=None, proxies=None, chunk_size=1024 * 50, *args, **kwargs):
        """
        解析财务文件

        :param report_hook: 钩子回调函数
        :param downdir: 要解析的文件夹
        :param proxies:
        :param chunk_size:
        :param args:
        :param kwargs:
        :return:
        """
        from pytdx.hq import TdxHq_API

        api = TdxHq_API(**kwargs)
        api.need_setup = False

        try:
            default = settings.get('SERVER').get('GP')[0][1:]
            bestip = config.get('BESTIP').get('GP', default)
        except ValueError:
            bestip = ("106.14.95.149", 7727)

        with api.connect(*bestip):
            content = api.get_report_file_by_size("tdxfin/gpcw.txt")

            if downdir is None:
                download_file = tempfile.NamedTemporaryFile(delete=True)
            else:
                download_file = open(downdir, 'wb')

            download_file.write(content)
            download_file.seek(0)

            return download_file

    def parse(self, download_file, *args, **kwargs):
        """
        解析财务文件

        :param download_file:
        :param args:
        :param kwargs:
        :return:
        """

        content = download_file.read()
        content = content.decode("utf-8")

        def list_to_dict(i):
            return {'filename': i[0], 'hash': i[1], 'filesize': int(i[2])}

        if content:
            return [list_to_dict(i) for i in [line.strip().split(",") for line in content.strip().split('\n')]]

        return None


class Financial(BaseFinancial):
    mode = "content"

    def build_url(self, *args, **kwargs):
        """
        获取采集数据的 URL

        :param args:
        :param kwargs:
        :return:
        """
        filename = kwargs.get('filename')

        if not filename:
            raise Exception("Param filename is not set")

        return f"http://down.tdx.com.cn:8001/fin/{filename}"

    def content(self, report_hook=None, downdir=None, proxies=None, chunk_size=1024 * 50, *args, **kwargs):
        """
        解析财务文件

        :param report_hook: 钩子回调函数
        :param downdir: 要解析的文件夹
        :param proxies:
        :param chunk_size:
        :param args:
        :param kwargs:
        :return:
        """
        filename = kwargs.get('filename')
        filesize = kwargs.get("filesize") if kwargs.get("filesize") else 0

        if not filename:
            raise Exception("Param filename is not set")

        api = TdxHq_API()
        api.need_setup = False

        try:
            default = settings.get('SERVER').get('GP')[0][1:]
            bestip = config.get('BESTIP').get('GP', default)
        except ValueError:
            bestip = ("106.14.95.149", 7727)

        with api.connect(*bestip):
            content = api.get_report_file_by_size(f"tdxfin/{filename}", filesize=filesize, reporthook=report_hook)
            download_file = open(downdir, 'wb') if downdir else tempfile.NamedTemporaryFile(delete=True)
            download_file.write(content)
            download_file.seek(0)

            return download_file

    def parse(self, download_file, *args, **kwargs):
        """
        解析财务文件

        :param download_file: 要解析的文件
        :param args:
        :param kwargs:
        :return:
        """

        header_pack_format = '<1hI1H3L'
        tmpdir = tempfile.gettempdir()

        if download_file.name.endswith('.zip'):
            tmpdir_root = tempfile.gettempdir()
            random_sums = str(random.randint(0, 1000000))
            subdir_name = f"mootdx_{random_sums}"

            tmpdir = Path(tmpdir_root, subdir_name)
            shutil.rmtree(tmpdir, ignore_errors=True)
            os.makedirs(tmpdir)

            shutil.unpack_archive(download_file.name, extract_dir=tmpdir)

            # only one file endswith .dat should be in zip archives
            datfile = None

            for _file in os.listdir(tmpdir):
                if _file.endswith(".dat"):
                    datfile = open(Path(tmpdir, _file), "rb")

            if datfile is None:
                raise Exception("no dat file found in zip archive")

        elif download_file.name.endswith('.dat'):
            datfile = download_file
        else:
            return None

        header_size = calcsize(header_pack_format)
        stock_item_size = calcsize("<6s1c1L")
        data_header = datfile.read(header_size)
        stock_header = unpack(header_pack_format, data_header)

        max_count = stock_header[2]

        report_date = stock_header[1]
        report_size = stock_header[4]

        report_fields_count = int(report_size / 4)
        report_pack_format = '<{}f'.format(report_fields_count)

        results = []

        for stock_idx in range(0, max_count):
            datfile.seek(header_size + stock_idx * calcsize("<6s1c1L"))
            si = datfile.read(stock_item_size)
            stock_item = unpack("<6s1c1L", si)
            code = stock_item[0].decode("utf-8")
            foa = stock_item[2]
            datfile.seek(foa)

            info_data = datfile.read(calcsize(report_pack_format))
            cw_info = unpack(report_pack_format, info_data)
            one_record = (code, report_date) + cw_info
            results.append(one_record)

        if download_file.name.endswith('.zip'):
            datfile.close()
            shutil.rmtree(tmpdir, ignore_errors=True)

        return results

    @staticmethod
    def to_df(data):
        """
        转换数据为 pandas DataFrame 格式
        :param data: 要转换的数据
        :return: DataFrame
        """

        if len(data) == 0:
            return None

        column = ['code', 'report_date']
        length = len(data[0]) - 1

        for i in range(1, length):
            column.append("col" + str(i))

        df = pd.DataFrame(data=data, columns=column)
        df.set_index('code', inplace=True)

        return df



# -*- coding: utf-8 -*-
import json
import logging
import socket
import time

from prettytable import PrettyTable

from mootdx import config
from mootdx.consts import CONFIG, EX_HOSTS, GP_HOSTS, HQ_HOSTS
from mootdx.logger import log
from mootdx.utils import get_config_path

result = []

hosts = {
    'HQ': [{
        'addr': hs[1],
        'port': hs[2],
        'time': 0,
        'site': hs[0]
    } for hs in HQ_HOSTS],
    'EX': [{
        'addr': hs[1],
        'port': hs[2],
        'time': 0,
        'site': hs[0]
    } for hs in EX_HOSTS],
    'GP': [{
        'addr': hs[1],
        'port': hs[2],
        'time': 0,
        'site': hs[0]
    } for hs in GP_HOSTS],
}


def Server(index=None, limit=5, console=False, verbose=False):
    if verbose:
        logging.basicConfig(level='DEBUG')

    _hosts = hosts[index]
    server = []

    while True:
        if len(_hosts) == 0:
            break

        try:
            proxy = _hosts.pop()
        except ValueError:
            return

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(5)

        try:
            start = time.perf_counter()

            sock.connect((proxy.get('addr'), int(proxy.get('port'))))
            sock.close()

            proxy['time'] = (time.perf_counter() - start) * 1000

            server.append(proxy)
            log.info("{}:{} 验证通过，响应时间：{:5.2f} ms.".format(proxy.get('addr'), proxy.get('port'), proxy.get('time')))
        except Exception as e:
            log.error(e)
            log.warning("{},{} 验证失败.".format(proxy.get('addr'), proxy.get('port')))

    # 结果按响应时间从小到大排序
    server.sort(key=lambda item: item['time'])

    if console:
        print("[√] 最优服务器:")

        t = PrettyTable(["Name", "Addr", "Port", "Time"])
        t.align["Name"] = "l"
        t.align["Addr"] = "l"
        t.align["Port"] = "l"
        t.align["Time"] = "r"
        t.padding_width = 1

        for host in server[:int(limit)]:
            t.add_row([
                host['site'], host['addr'], host['port'],
                '{:5.2f} ms'.format(host['time'])
            ])

        print(t)

    return [(item['addr'], item['port']) for item in server]


def bestip():
    config_ = get_config_path('config.json')
    default = dict(CONFIG)

    for index in ['HQ', 'EX', 'GP']:
        result = Server(index=index)

        if result:
            default['BESTIP'][index] = result[0]

    json.dump(default, open(config_, 'w'), indent=2)
    config.setup()



import socket
import threading

from pytdx.base_socket_client import (CONNECT_TIMEOUT, BaseSocketClient,
                                      TrafficStatSocket)
from pytdx.errors import TdxConnectionError
from pytdx.heartbeat import HqHeartBeatThread
from pytdx.reader import TdxDailyBarReader

from mootdx.logger import log


class MooTdxDailyBarReader(TdxDailyBarReader):
    # 感谢 bopomofo 的鼎力支持
    SECURITY_TYPE = ["SH_A_STOCK", "SH_B_STOCK", "SH_STAR_STOCK", "SH_INDEX", "SH_FUND", "SH_BOND", "SZ_A_STOCK", "SZ_B_STOCK", "SZ_INDEX", "SZ_FUND", "SZ_BOND"]
    SECURITY_COEFFICIENT = {"SH_A_STOCK": [0.01, 0.01], "SH_B_STOCK": [0.001, 0.01], "SH_STAR_STOCK": [0.01, 0.01], "SH_INDEX": [0.01, 1.0], "SH_FUND": [0.001, 1.0],
                            "SH_BOND": [0.001, 1.0], "SZ_A_STOCK": [0.01, 0.01], "SZ_B_STOCK": [0.01, 0.01], "SZ_INDEX": [0.01, 1.0], "SZ_FUND": [0.001, 0.01],
                            "SZ_BOND": [0.001, 0.01]}

    def get_security_type(self, fname):

        exchange = str(fname[-12:-10]).lower()
        code_head = fname[-10:-8]

        if exchange == self.SECURITY_EXCHANGE[0]:
            if code_head in ["00", "30"]:
                return "SZ_A_STOCK"
            elif code_head in ["20"]:
                return "SZ_B_STOCK"
            elif code_head in ["39"]:
                return "SZ_INDEX"
            elif code_head in ["15", "16"]:
                return "SZ_FUND"
            elif code_head in ["10", "11", "12", "13", "14"]:
                return "SZ_BOND"
        elif exchange == self.SECURITY_EXCHANGE[1]:
            if code_head in ["60"]:
                return "SH_A_STOCK"
            elif code_head in ["90"]:
                return "SH_B_STOCK"
            elif code_head in ["68"]:
                return "SH_STAR_STOCK"
            elif code_head in ["00", "88", "99"]:
                return "SH_INDEX"
            elif code_head in ["50", "51"]:
                return "SH_FUND"
            elif code_head in ["01", "10", "11", "12", "13", "14"]:
                return "SH_BOND"
        else:
            log.error("Unknown security exchange !\n")
            raise NotImplementedError


class MooBaseSocketClient(BaseSocketClient):

    def connect(self, ip='101.227.73.20', port=7709, time_out=CONNECT_TIMEOUT, bindport=None, bindip='0.0.0.0'):
        """

        :param ip:  服务器ip 地址
        :param port:  服务器端口
        :param time_out: 连接超时时间
        :param bindport: 绑定的本地端口
        :param bindip: 绑定的本地ip
        :return: 是否连接成功 True/False
        """

        self.client = TrafficStatSocket(socket.AF_INET, socket.SOCK_STREAM)
        self.client.settimeout(time_out)

        log.debug("connecting to server : %s on port :%d" % (ip, port))

        try:
            self.ip = ip
            self.port = port

            if bindport is not None:
                self.client.bind((bindip, bindport))

            self.client.connect((ip, port))
        except socket.timeout as e:
            log.debug(e)
            log.debug("connection expired")

            if self.raise_exception:
                raise TdxConnectionError("connection timeout error")

            return False
        except Exception as e:
            log.debug(e)
            if self.raise_exception:
                raise TdxConnectionError("other errors")

            return False

        log.debug("connected!")

        if self.need_setup:
            self.setup()

        if self.heartbeat:
            self.stop_event = threading.Event()
            self.heartbeat_thread = HqHeartBeatThread(self, self.stop_event, self.heartbeat_interval)
            self.heartbeat_thread.start()

        return self

    def disconnect(self):

        if self.heartbeat_thread and self.heartbeat_thread.is_alive():
            self.stop_event.set()

        if self.client:
            log.debug("disconnecting")

            try:
                self.client.shutdown(socket.SHUT_RDWR)
                self.client.close()
                self.client = None
            except Exception as e:
                log.debug(str(e))

                if self.raise_exception:
                    raise TdxConnectionError("disconnect err")

            log.debug("disconnected")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""The setup script."""

try:
    from setuptools import find_packages, setup
except ImportError:
    from distutils.core import find_packages, setup


def parse_requirements(filename):
    line_iter = (line.strip() for line in open(filename))
    return [line for line in line_iter if line and not line.startswith("#")]


with open('README.rst', encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst', encoding='utf-8') as history_file:
    history = history_file.read()

requirements = parse_requirements('requirements.txt')
test_requirements = requirements
test_requirements.append('pytest')

setup(
    name='mootdx',
    version='0.7.5',
    description="通达信数据读取接口.",
    long_description=readme,
    author="bopo.wang",
    author_email='ibopo@126.com',
    url='https://github.com/bopo/mootdx',
    packages=find_packages(include=['mootdx', 'mootdx.*']),
    # include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='mootdx',
    entry_points={'console_scripts': [
        'mootdx=mootdx.__main__:entry',
    ]},
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=requirements,
)



# -*- coding: utf-8 -*-
from abc import ABC

from pytdx.reader import (BlockReader, CustomerBlockReader, TdxExHqDailyBarReader, TdxLCMinBarReader)
from unipath import Path

from mootdx import utils
from mootdx.consts import TYPE_GROUP
from mootdx.contrib.compat import MooTdxDailyBarReader
from mootdx.logger import log
from mootdx.utils import get_stock_market


# 股票市场
class Reader(object):
    @staticmethod
    def factory(market='std', **kwargs):
        """
        Reader 工厂方法

        :param market:  std 标准市场, ext 扩展市场
        :param kwargs:
        :return:
        """
        if market == 'ext':
            return ExtReader(**kwargs)
        elif market == 'std':
            return StdReader(**kwargs)


class ReaderBase(ABC):
    """股票市场"""

    # 默认通达信安装目录
    tdxdir = 'C:/new_tdx'

    def __init__(self, tdxdir=None):
        """
        构造函数

        :param tdxdir: 通达信安装目录
        """

        if Path(tdxdir).isdir():
            self.tdxdir = tdxdir
        else:
            log.error('tdxdir 目录不存在')
            raise Exception('tdxdir 目录不存在')

    def find_path(self, symbol=None, subdir='lday', suffix=None):
        """
        自动匹配文件路径，辅助函数

        :param symbol:
        :param subdir:
        :param suffix:
        :return: pd.dataFrame or None
        """
        market = get_stock_market(symbol, True) if len(symbol.split('#')) == 1 else 'ds'
        prefix = market if len(symbol.split('#')) == 1 else ''
        suffix = suffix if isinstance(suffix, list) else [suffix]

        for ex_ in suffix:
            ex_ = ex_.strip('.')
            vipdoc = Path(self.tdxdir, 'vipdoc', market, subdir, f'{prefix}{symbol}.{ex_}')

            if not Path(vipdoc).exists():
                log.debug(f'未找到所需的文件: {vipdoc}')
                continue

            log.debug(f"找到所需的文件: {vipdoc}")
            return vipdoc

        return None


class StdReader(ReaderBase):
    """股票市场"""

    def daily(self, symbol=None):
        """
        获取日线数据

        :param symbol:
        :return: pd.dataFrame or None
        """
        reader = MooTdxDailyBarReader()
        vipdoc = self.find_path(symbol=symbol, subdir='lday', suffix='day')

        if vipdoc is not None:
            return reader.get_df(vipdoc)

        return None

    def minute(self, symbol=None, suffix=1):
        """
        获取1, 5分钟线

        :param suffix:
        :param symbol:
        :return: pd.dataFrame or None
        """
        subdir = 'fzline' if str(suffix) == '5' else 'minline'
        suffix = ['lc5', '5'] if str(suffix) == '5' else ['lc1', '1']
        symbol = self.find_path(symbol, subdir=subdir, suffix=suffix)
        reader = TdxLCMinBarReader()

        if symbol is not None:
            return reader.get_df(symbol)

        return None

    def fzline(self, symbol=None):
        """
        分钟线数据

        :param symbol: 自定义板块股票列表, 类型 list
        :return: pd.dataFrame or Bool
        """
        return self.minute(symbol, suffix=5)

    def block_new(self, name: str = None, symbol: list = None, group=False):
        """
        自定义板块数据操作
        提示: name 和 symbol 全为空则为读取，否则写入操作
        参考: http://blog.sina.com.cn/s/blog_623d2d280102vt8y.html

        :param name: 自定义板块名称
        :param symbol: 自定义板块股票列表, 类型 list
        :param group:
        :return: pd.dataFrame or Bool
        """

        if name or symbol:
            return utils.block_new(self.tdxdir, name=name, symbol=symbol)

        reader = CustomerBlockReader()
        vipdoc = Path(self.tdxdir, 'T0002', 'blocknew')

        fmt = TYPE_GROUP if group else None

        if Path(vipdoc).isdir():
            log.debug(f'找到所需的文件: {vipdoc}')
            return reader.get_df(vipdoc, fmt)

        log.error(f'未找到所需的文件: {vipdoc}')

        return None

    def block(self, symbol='', custom=False, group=False):
        """
        获取板块数据
        参考: http://blog.sina.com.cn/s/blog_623d2d280102vt8y.html

        :param custom:
        :param symbol:
        :param group:
        :return: pd.dataFrame or None
        """

        # suffix = symbol.split('.')
        # suffix = suffix[-1] if len(suffix) > 1 else 'dat'
        suffix = 'dat'
        # symbol = suffix[0]

        reader = BlockReader()
        vipdoc = Path(self.tdxdir, 'T0002', 'hq_cache', f'{symbol}.{suffix}')

        fmt = TYPE_GROUP if group else None

        if Path(vipdoc).exists():
            return reader.get_df(vipdoc, fmt)

        log.error(f'未找到所需的文件: {vipdoc}')

        return None


class ExtReader(ReaderBase):
    """扩展市场读取"""

    def __init__(self, tdxdir=None):
        super(ExtReader, self).__init__(tdxdir)
        self.reader = TdxExHqDailyBarReader()

    def daily(self, symbol=None):
        """
        获取扩展市场日线数据

        :return: pd.dataFrame or None
        """
        vipdoc = self.find_path(symbol=symbol, subdir='lday', suffix='day')

        if symbol is not None:
            return self.reader.get_df(vipdoc)

        return None

    def minute(self, symbol=None):
        """
        获取扩展市场分钟线数据

        :return: pd.dataFrame or None
        """
        vipdoc = self.find_path(symbol=symbol, subdir='minline', suffix='lc1')

        if symbol is not None:
            return self.reader.get_df(vipdoc)

        return None

    def fzline(self, symbol=None):
        """
        获取日线数据

        :return: pd.dataFrame or None
        """
        vipdoc = self.find_path(symbol=symbol, subdir='fzline', suffix='lc5')

        if symbol is not None:
            return self.reader.get_df(vipdoc)

        return None



# -*- coding: utf-8 -*-
import contextlib
import math

import pandas
from pytdx.exhq import TdxExHq_API
from pytdx.hq import TdxHq_API
from tqdm import tqdm
from unipath import Path

from mootdx import config, server
from mootdx.consts import MARKET_SH
from mootdx.logger import log
from mootdx.utils import (get_config_path, get_stock_market, get_stock_markets, to_data)


class Quotes(object):
    @staticmethod
    def factory(market='std', **kwargs):
        """
        股票市场 工厂方法

        :param market: std 标准市场, ext 扩展市场
        :param kwargs:
        :return: object
        """
        if market == 'ext':
            return ExtQuotes(**kwargs)
        elif market == 'std':
            return StdQuotes(**kwargs)


class BaseQuotes(object):
    client = None
    bestip = None

    def __init__(self, bestip=False, timeout=15, **kwargs):
        log.debug(f'bestip=>{bestip}')

        self.timeout = timeout

        config_ = get_config_path('config.json')

        if bestip or not Path(config_).exists():
            server.bestip()

    @contextlib.contextmanager
    def connect(self):
        if self.closed:
            log.debug('服务器连接已断开，正进行重新连接...')
            self.reconnect()

        yield

        self.close()

    def __del__(self):
        log.debug('__del__')
        self.close()

    def reconnect(self):
        if self.closed:
            log.debug('服务器连接已断开，正进行重新连接...')
            self.client.connect(time_out=self.timeout, *self.bestip)

    def close(self):
        log.debug('close')
        self.client.close()

    @property
    def closed(self):
        if hasattr(self.client.client, '_closed') and not getattr(self.client.client, '_closed'):
            return False

        return True


class StdQuotes(BaseQuotes):
    """
    股票市场实时行情
    """
    bestip = ('47.103.48.45', 7709)

    def __init__(self, bestip=False, timeout=15, **kwargs):
        super(StdQuotes, self).__init__(bestip=bestip, timeout=timeout, **kwargs)

        try:
            default = config.get('SERVER').get('HQ')[0]
            self.bestip = config.get('BESTIP').get('HQ', default)
        except ValueError:
            self.bestip = ('47.103.48.45', 7709)

        self.client = TdxHq_API(**kwargs)
        self.client.connect(*self.bestip)

    def traffic(self):
        self.reconnect()
        return self.client.get_traffic_stats()

    def quotes(self, symbol=None):
        """
        获取实时日行情数据

        :param symbol: 股票代码
        :return: pd.dataFrame or None
        """

        if not symbol:
            return to_data(None)

        if type(symbol) is str:
            symbol = [symbol]

        self.reconnect()

        symbol = get_stock_markets(symbol)
        result = self.client.get_security_quotes(symbol)

        return to_data(result)

    def bars(self, symbol='000001', frequency='9', start='0', offset='100', *args, **kwargs):
        """
        获取实时日K线数据

        :param symbol: 股票代码
        :param frequency: 数据类别
        :param start: 开始位置
        :param offset: 每次获取条数
        :return: pd.dataFrame or None
        """
        self.reconnect()

        market = get_stock_market(symbol)
        result = self.client.get_security_bars(int(frequency), int(market), str(symbol), int(start), int(offset))

        return to_data(result)

    def stock_count(self, market=MARKET_SH, *args, **kwargs):
        """
        获取市场股票数量

        :param market: 股票市场代码 sh 上海， sz 深圳
        :return: pd.dataFrame or None
        """
        self.reconnect()

        result = self.client.get_security_count(market=market)

        return result

    def stocks(self, market=MARKET_SH, *args, **kwargs):
        """
        获取股票列表

        :param market: 股票市场
        :return:
        """
        self.reconnect()

        counts = self.client.get_security_count(market=market)
        stocks = None

        for start in tqdm(range(0, counts, 1000)):
            result = self.client.get_security_list(market=market, start=start)
            stocks = pandas.concat([stocks, to_data(result)], ignore_index=True) if start > 1 else to_data(result)

        return stocks

    def index_bars(self, symbol='000001', frequency='9', start='0', offset='100', *args, **kwargs):
        """
        获取指数k线

        :param symbol: 股票代码
        :param frequency:
        :param start:
        :param offset:
        :return:
        """
        self.reconnect()

        market = get_stock_market(symbol)
        result = self.client.get_index_bars(frequency=frequency, market=market, code=symbol, start=start, count=offset)

        return to_data(result)

    def minute(self, symbol=''):
        """
        获取实时分时数据

        :param symbol: 股票代码
        :return: pd.DataFrame
        """
        self.reconnect()

        market = get_stock_market(symbol)
        result = self.client.get_minute_time_data(market=market, code=symbol)

        return to_data(result)

    def minutes(self, symbol='', date='20191023', *args, **kwargs):
        """
        分时历史数据

        :param symbol: 股票代码
        :param date:
        :return: pd.dataFrame or None
        """
        self.reconnect()

        market = get_stock_market(symbol)
        result = self.client.get_history_minute_time_data(market=market, code=symbol, date=date)

        return to_data(result)

    def transaction(self, symbol='', start=0, *args, **kwargs):
        """
        查询分笔成交

        :param symbol: 股票代码
        :param start: 起始位置
        :return: pd.dataFrame or None
        """
        self.reconnect()

        market = get_stock_market(symbol)
        result = self.client.get_transaction_data(int(market), symbol, int(start), int(market))

        return to_data(result)

    def transactions(self, symbol='', start=0, offset=10, date='20170209'):
        """
        查询历史分笔成交
        参数：市场代码， 股票代码，起始位置，日期 数量 如： 0,000001,0,10,20170209

        :param symbol: 股票代码
        :param start: 起始位置
        :param offset: 数量
        :param date: 日期
        :return: pd.dataFrame or None
        """
        self.reconnect()

        market = get_stock_market(symbol, string=False)
        result = self.client.get_history_transaction_data(market=market, code=symbol, start=start, count=offset, date=int(date))

        return to_data(result)

    def F10C(self, symbol=''):
        """
        查询公司信息目录

        :param symbol: 股票代码
        :return: pd.dataFrame or None
        """
        self.reconnect()

        market = int(get_stock_market(symbol))
        result = self.client.get_company_info_category(market, symbol)

        return result

    def F10(self, symbol='', name=''):
        """
        读取公司信息详情

        :param name: 公司 F10 标题
        :param symbol: 股票代码
        :return: pd.dataFrame or None
        """
        self.reconnect()

        result = {}
        market = int(get_stock_market(symbol, string=False))

        frequency = self.client.get_company_info_category(market, symbol)

        if name:
            for x in frequency:
                if x['name'] == name:
                    return self.client.get_company_info_content(market=market, code=symbol, filename=x['filename'], start=x['start'], length=x['length'])

        for x in frequency:
            result[x['name']] = self.client.get_company_info_content(
                market=market, code=symbol, filename=x['filename'], start=x['start'], length=x['length'])

        return result

    def xdxr(self, symbol=''):
        """
        读取除权除息信息

        :param symbol: 股票代码
        :return: pd.dataFrame or None
        """
        self.reconnect()

        market = get_stock_market(symbol)
        result = self.client.get_xdxr_info(int(market), symbol)

        return to_data(result)

    def finance(self, symbol='000001'):
        """
        读取财务信息

        :param symbol: 股票代码
        :return:
        """
        self.reconnect()

        market = get_stock_market(symbol)
        result = self.client.get_finance_info(market=market, code=symbol)

        return to_data(result)

    def k(self, symbol='', begin=None, end=None):
        """
        读取k线信息

        :param symbol:
        :param begin: 开始日期
        :param end: 截止日期
        :return: pd.dataFrame or None
        """
        self.reconnect()

        result = self.client.get_k_data(symbol, begin, end)
        return result

    def index(self, symbol='000001', market=MARKET_SH, frequency='9', start=1, offset=2):
        """
        获取指数k线

        K线种类:
        - 0 5分钟K线
        - 1 15分钟K线
        - 2 30分钟K线
        - 3 1小时K线
        - 4 日K线
        - 5 周K线
        - 6 月K线

        - 7 1分钟
        - 8 1分钟K线
        - 9 日K线
        - 10 季K线
        - 11 年K线

        :param symbol: 股票代码
        :param frequency: 数据类别
        :param market: 证券市场
        :param start: 开始位置
        :param offset: 每次获取条数
        :return: pd.dataFrame or None
        """
        self.reconnect()

        result = self.client.get_index_bars(int(frequency), int(market), str(symbol), int(start), int(offset))
        return to_data(result)

    def block(self, tofile="block.dat"):
        """
        获取证券板块信息

        :param tofile:
        :return: pd.dataFrame or None
        """
        self.reconnect()

        result = self.client.get_and_parse_block_info(tofile)
        return to_data(result)


class ExtQuotes(BaseQuotes):
    """
    扩展市场实时行情
    """

    bestip = ('112.74.214.43', 7727)

    def __init__(self, bestip=False, timeout=15, **kwargs):
        super(ExtQuotes, self).__init__(bestip=bestip, timeout=timeout, **kwargs)

        try:
            default = config.get('SERVER').get('EX')[0]
            self.bestip = config.get('BESTIP').get('EX', default)
        except ValueError:
            self.bestip = ('112.74.214.43', 7727)

        log.debug(self.bestip)

        self.client = TdxExHq_API(**kwargs)
        self.client.connect(*self.bestip)

    @staticmethod
    def validate(market, symbol):
        if not market:
            if len(symbol.split('#')) > 1:
                market = symbol.split('#')[0]
                symbol = symbol.split('#')[1]

        if not market:
            raise ValueError('市场参数错误, 市场参数不能为空.')

        return int(market), symbol

    def markets(self):
        """
        获取实时市场列表

        :return: pd.dataFrame or None
        """
        self.reconnect()

        result = self.client.get_markets()
        return to_data(result)

    def instrument(self, start=0, offset=100):
        """
        查询代码列表

        :param offset:
        :param start:
        :return:
        """
        self.reconnect()

        result = self.client.get_instrument_info(start=start, count=offset)
        return to_data(result)

    def instrument_count(self):
        """
        市场商品数量

        :return:
        """
        self.reconnect()

        result = self.client.get_instrument_count()

        return result

    def instruments(self):
        """
        查询所有代码列表

        :return:
        """
        self.reconnect()

        result = []

        count = self.client.get_instrument_count()
        pages = math.ceil(count / 100)

        for page in tqdm(range(0, pages)):
            result += self.client.get_instrument_info(page * 100, 100)

        return to_data(result)

    def quote(self, market='', symbol=''):
        """
        查询五档行情

        :param market: 市场ID
        :param symbol: 证券代码
        :return:
        """
        self.reconnect()

        market, symbol = self.validate(market, symbol)
        result = self.client.get_instrument_quote(market, symbol)

        return to_data(result)

    def minute(self, market='', symbol=''):
        """
        查询分时行情

        :param market: 市场ID
        :param symbol: 证券代码
        :return:
        """
        self.reconnect()

        market, symbol = self.validate(market, symbol)
        result = self.client.get_minute_time_data(market, symbol)

        return to_data(result)

    def minutes(self, market=None, symbol='', date=''):
        """
        查询历史分时行情

        :param market: 市场ID
        :param symbol: 证券代码
        :param date:
        :return:
        """
        self.reconnect()

        market, symbol = self.validate(market, symbol)
        result = self.client.get_history_minute_time_data(market, symbol, date)

        return to_data(result)

    def bars(self, frequency='', market='', symbol='', start=0, offset=100):
        """
        查询k线数据
        参数： K线周期， 市场ID， 证券代码，起始位置， 数量

        :param frequency: K线周期
        :param market: 市场ID
        :param symbol: 证券代码
        :param start: 起始位置
        :param offset: 数量
        :return:
        """
        self.reconnect()

        market, symbol = self.validate(market, symbol)
        result = self.client.get_instrument_bars(category=frequency, market=market, code=symbol, start=start, count=offset)

        return to_data(result)

    def transaction(self, market=None, symbol='', start=0, offset=1800):
        """
        查询分笔成交

        :param market: 市场ID
        :param symbol: 证券代码
        :param start:
        :param offset:
        :return:
        """
        self.reconnect()

        market, symbol = self.validate(market, symbol)
        result = self.client.get_transaction_data(market=market, code=symbol, start=start, count=offset)

        return to_data(result)

    def transactions(self, market=None, symbol='', date='', start=0, offset=1800):
        """
        查询历史分笔成交

        :param market: 市场ID
        :param symbol: 证券代码
        :param date:
        :param start:
        :param offset:
        :return:
        """
        self.reconnect()

        market, symbol = self.validate(market, symbol)
        result = self.client.get_history_transaction_data(market=market, code=symbol, date=int(date), start=start, count=offset)

        return to_data(result)
